__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


from copy import deepcopy
from itertools import product

import pytest

from wisdem.orbit.phases.design import SparDesign

base = {
    "site": {"depth": 500},
    "plant": {"num_turbines": 50},
    "turbine": {"turbine_rating": 12},
    "spar_design": {},
}


@pytest.mark.parametrize(
    "depth,turbine_rating",
    product(range(100, 1201, 200), range(3, 15, 1)),
)
def test_parameter_sweep(depth, turbine_rating):

    config = {
        "site": {"depth": depth},
        "plant": {"num_turbines": 50},
        "turbine": {"turbine_rating": turbine_rating},
        "substation_design": {},
    }

    s = SparDesign(config)
    s.run()

    assert s.detailed_output["stiffened_column_mass"] > 0
    assert s.detailed_output["tapered_column_mass"] > 0
    assert s.detailed_output["ballast_mass"] > 0
    assert s.detailed_output["secondary_steel_mass"] > 0


def test_design_kwargs():

    test_kwargs = {
        "stiffened_column_CR": 3000,
        "tapered_column_CR": 4000,
        "ballast_material_CR": 200,
        "secondary_steel_CR": 7000,
    }

    s = SparDesign(base)
    s.run()
    base_cost = s.total_cost

    for k, v in test_kwargs.items():

        config = deepcopy(base)
        config["spar_design"] = {}
        config["spar_design"][k] = v

        s = SparDesign(config)
        s.run()
        cost = s.total_cost

        assert cost != base_cost


def test_total_cost():

    spar = SparDesign(base)
    spar.run()

    assert spar.total_cost == pytest.approx(698569358, abs=1e0)



"""Initializes ORBIT and provides the top-level import objects."""

__author__ = [
    "Jake Nunemaker",
    "Matt Shields",
    "Rob Hammond",
    "Nick Riccobono",
]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Nick Riccobono"
__email__ = ["nicholas.riccobono@nrel.gov", "rob.hammond@nrel.gov"]
__status__ = "Development"


from wisdem.orbit.manager import ProjectManager  # isort:skip
from wisdem.orbit.config import load_config, save_config
from wisdem.orbit.parametric import ParametricManager
from wisdem.orbit.supply_chain import SupplyChainManager

__version__ = "1.1"



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

from pathlib import Path

import pandas as pd

DIR = Path(__file__).resolve().parent
_fp = DIR / "test_weather.csv"
test_weather = (
    pd.read_csv(_fp, parse_dates=["datetime"])
    .set_index("datetime")
    .to_records()
)



"""Provides the ``Cargo`` base class."""

from marmot import Object


class Cargo(Object):
    """Base class for working with cargo."""

    def __repr__(self):
        """Overridden __repr__ method."""
        return self.type

    @property
    def type(self):
        """Returns type of `Cargo`."""
        return self.__class__.__name__



__author__ = "Jake Nunemaker, Sophie Bredenkamp"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


import warnings
from copy import deepcopy
from itertools import product

import pytest

from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.phases.design import ElectricalDesign, OffshoreSubstationDesign

# OSS TESTING

base = {
    "site": {"distance_to_landfall": 50, "depth": 30},
    "plant": {"capacity": 500},
    "export_system_design": {"cables": "XLPE_630mm_220kV"},
    "landfall": {},
    "substation_design": {
        "oss_pile_cost_rate": 1200,  # need to set this for kwarg tests
    },
}


@pytest.mark.parametrize(
    "distance_to_landfall,depth,plant_cap,cable",
    product(
        range(10, 201, 50),
        range(10, 51, 10),
        range(100, 2001, 500),
        ["XLPE_630mm_220kV", "XLPE_800mm_220kV", "XLPE_1000mm_220kV"],
    ),
)
def test_parameter_sweep(distance_to_landfall, depth, plant_cap, cable):
    config = {
        "site": {"distance_to_landfall": distance_to_landfall, "depth": depth},
        "plant": {"capacity": plant_cap},
        "export_system_design": {"cables": cable},
        "substation_design": {},
    }

    elect = ElectricalDesign(config)
    elect.run()

    # Check valid substructure length
    assert (
        10
        <= elect._outputs["offshore_substation_substructure"]["length"]
        <= 80
    )

    # Check valid substructure mass
    assert (
        200
        <= elect._outputs["offshore_substation_substructure"]["mass"]
        <= 2700
    )

    # Check valid topside mass
    assert 200 <= elect._outputs["offshore_substation_topside"]["mass"] <= 5500

    # Check valid substation cost
    assert 1e6 <= elect.total_substation_cost <= 1e9


def test_detailed_design_length():
    """Ensure that the same # of output variables are used for a floating
    and fixed offshore substation.
    """

    elect = ElectricalDesign(base)
    elect.run()

    floating = deepcopy(base)
    floating["substation_design"]["oss_substructure_type"] = "Floating"
    elect_floating = ElectricalDesign(floating)
    elect_floating.run()

    assert len(elect.detailed_output) == len(elect_floating.detailed_output)


def test_calc_substructure_mass_and_cost():

    elect = ElectricalDesign(base)
    elect.run()

    floating = deepcopy(base)
    floating["substation_design"]["oss_substructure_type"] = "Floating"
    elect_floating = ElectricalDesign(floating)
    elect_floating.run()

    assert (
        elect.detailed_output["substation_substructure_cost"]
        != elect_floating.detailed_output["substation_substructure_cost"]
    )
    assert (
        elect.detailed_output["substation_substructure_mass"]
        != elect_floating.detailed_output["substation_substructure_mass"]
    )


def test_calc_topside_mass_and_cost():
    # Test topside mass and cost for HVDC compared to HVDC-Monopole and
    # HVDC-Bipole.
    #

    config = deepcopy(base)
    config["substation_design"]["topside_design_cost"] = 9999.9
    elect = ElectricalDesign(config)
    elect.run()

    assert elect._outputs["num_substations"] == 1
    assert elect._outputs["offshore_substation_topside"][
        "unit_cost"
    ] == pytest.approx(23683541, abs=1e2)

    mono_dc = deepcopy(base)
    mono_dc["export_system_design"]["cables"] = "HVDC_2000mm_320kV"
    elect_mono = ElectricalDesign(mono_dc)
    elect_mono.run()

    assert (
        elect.detailed_output["substation_topside_mass"]
        == elect_mono.detailed_output["substation_topside_mass"]
    )
    assert (
        elect.detailed_output["substation_topside_cost"]
        != elect_mono.detailed_output["substation_topside_cost"]
    )

    bi_dc = deepcopy(base)
    bi_dc["export_system_design"]["cables"] = "HVDC_2500mm_525kV"
    elect_bi = ElectricalDesign(bi_dc)
    elect_bi.run()

    assert (
        elect.detailed_output["substation_topside_mass"]
        == elect_bi.detailed_output["substation_topside_mass"]
    )
    assert (
        elect.detailed_output["substation_topside_cost"]
        != elect_bi.detailed_output["substation_topside_cost"]
    )

    assert (
        elect_bi.detailed_output["substation_topside_mass"]
        == elect_mono.detailed_output["substation_topside_mass"]
    )
    assert (
        elect_bi.detailed_output["substation_topside_cost"]
        != elect_mono.detailed_output["substation_topside_cost"]
    )


def test_oss_substructure_kwargs():
    test_kwargs = {
        "oss_substructure_type": "Floating",
        "oss_substructure_cost_rate": 7250,
        "oss_pile_cost_rate": 2500,
        "num_substations": 4,
    }

    elect = ElectricalDesign(base)
    elect.run()
    base_cost_total = elect.detailed_output["total_substation_cost"]
    base_cost_subst = elect.detailed_output["substation_substructure_cost"]

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["substation_design"] = {}
        config["substation_design"][k] = v

        elect = ElectricalDesign(config)
        elect.run()
        cost_total = elect.detailed_output["total_substation_cost"]
        cost_subst = elect.detailed_output["substation_substructure_cost"]

        assert cost_total != base_cost_total
        assert cost_subst != base_cost_subst


def test_ac_oss_kwargs():
    test_kwargs = {
        "mpt_unit_cost": 13500,
        # "topside_fab_cost_rate": 17000,    # breaks
        "topside_design_cost": 7e6,
        "shunt_unit_cost": 40000,
        "switchgear_cost": 15e5,
        "backup_gen_cost": 2e6,
        "workspace_cost": 3e6,
        "other_ancillary_cost": 4e6,
        "topside_assembly_factor": 0.09,
        "num_substations": 4,
    }

    elect = ElectricalDesign(base)
    elect.run()
    base_cost = elect.detailed_output["total_substation_cost"]

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["substation_design"] = {}
        config["substation_design"][k] = v

        elect = ElectricalDesign(config)
        elect.run()
        cost = elect.detailed_output["total_substation_cost"]
        print("passed")
        assert cost != base_cost


def test_dc_oss_kwargs():
    test_kwargs = {"converter_cost": 300e6, "dc_breaker_cost": 300e6}

    dc_base = deepcopy(base)
    dc_base["export_system_design"]["cables"] = "HVDC_2000mm_320kV"
    elect = ElectricalDesign(dc_base)
    elect.run()
    base_cost = elect.detailed_output["total_substation_cost"]

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["export_system_design"]["cables"] = "HVDC_2000mm_320kV"
        config["substation_design"] = {}
        config["substation_design"][k] = v

        elect = ElectricalDesign(config)
        elect.run()
        cost = elect.detailed_output["total_substation_cost"]

        assert cost != base_cost


def test_new_old_hvac_substation():
    """Temporary test until ElectricalDesign is merged with new release."""

    config = deepcopy(base)
    config["export_system_design"] = {"cables": "HVDC_2000mm_320kV"}
    config["plant"]["capacity"] = 1000  # MW
    config["plant"]["num_turbines"] = 200
    config["turbine"] = {"turbine_rating": 5}

    config["export_system"] = {"cable": {"number": 5, "cable_type": "HVAC"}}

    new = ElectricalDesign(config)
    new.run()

    old = OffshoreSubstationDesign(config)
    old.run()

    # same values
    assert new.num_substations == old.num_substations
    assert new.topside_mass == old.topside_mass
    assert new.ancillary_system_costs == old.ancillary_system_costs
    assert new.substructure_mass == old.substructure_mass

    # different values
    assert new.substation_cost != old.substation_cost
    assert new.mpt_rating != old.mpt_rating
    assert new.num_mpt != old.num_mpt
    assert new.mpt_cost != old.mpt_cost
    assert new.topside_cost != old.topside_cost
    assert new.shunt_reactor_cost != old.shunt_reactor_cost


def test_hvac_substation():
    config = deepcopy(base)

    hvac = ElectricalDesign(config)
    hvac.run()

    assert hvac.total_substation_cost == pytest.approx(134448256, abs=1e0)


def test_hvdc_substation():
    config = deepcopy(base)
    config["export_system_design"] = {"cables": "HVDC_2000mm_320kV"}
    elect = ElectricalDesign(config)
    elect.run()

    assert elect.total_substation_cost == pytest.approx(451924714, abs=1e0)

    assert elect.converter_cost != 0
    assert elect.shunt_reactor_cost == 0
    assert elect.dc_breaker_cost != 0
    assert elect.switchgear_cost == 0
    assert elect.mpt_cost == 0
    # assert elect.num_cables / elect.num_converters == 2  # breaks

    config = deepcopy(base)
    config["export_system_design"] = {"cables": "HVDC_2500mm_525kV"}

    elect = ElectricalDesign(config)
    elect.run()

    assert elect.total_substation_cost == pytest.approx(802924714, abs=1e0)

    assert elect.converter_cost != 0
    assert elect.shunt_reactor_cost == 0
    assert elect.dc_breaker_cost != 0
    assert elect.switchgear_cost == 0
    assert elect.mpt_cost == 0
    # assert elect.num_cables / elect.num_converters == 2  # breaks


def test_onshore_substation():
    config = deepcopy(base)
    elect = ElectricalDesign(config)
    elect.run()
    assert elect.onshore_compensation_cost != 0.0
    assert elect.onshore_cost == pytest.approx(95.487e6, abs=1e2)  # 109.32e6

    config_mono = deepcopy(config)
    config_mono["export_system_design"] = {"cables": "HVDC_2000mm_320kV"}
    o_monelect = ElectricalDesign(config_mono)
    o_monelect.run()
    assert o_monelect.onshore_compensation_cost == 0.0
    assert o_monelect.onshore_cost == 244.3e6

    config_bi = deepcopy(config)
    config_bi["export_system_design"] = {"cables": "HVDC_2500mm_525kV"}
    o_bi = ElectricalDesign(config_bi)
    o_bi.run()
    assert o_bi.onshore_compensation_cost == 0.0
    assert o_bi.onshore_cost == 450e6


# EXPORT CABLE TESTING


def test_export_kwargs():
    test_kwargs = {
        "num_redundant": 2,
        "touchdown_distance": 50,
        "percent_added_length": 0.15,
        # "interconnection_distance": 6,
    }

    elect = ElectricalDesign(base)
    elect.run()
    base_cost = elect.total_cost

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["export_system_design"] = {"cables": "XLPE_630mm_220kV"}
        config["export_system_design"][k] = v

        elect = ElectricalDesign(config)
        elect.run()
        cost = elect.total_cost

        assert cost != base_cost


config = extract_library_specs("config", "export_design")


def test_export_system_creation():
    export = ElectricalDesign(config)
    export.run()

    assert isinstance(export.num_cables, int)
    assert export.length
    assert export.mass
    assert export.cable
    assert export.total_length
    assert export.total_mass
    assert export.num_substations
    assert export.topside_mass
    assert export.substructure_mass


def test_number_cables():
    export = ElectricalDesign(config)
    export.run()

    assert export.num_cables == 9


def test_cable_length():
    export = ElectricalDesign(config)
    export.run()

    length = (0.02 + 3 + 30) * 1.01
    assert export.length == length


def test_cable_mass():
    export = ElectricalDesign(config)
    export.run()

    length = (0.02 + 3 + 30) * 1.01
    mass = length * export.cable.linear_density
    assert export.mass == pytest.approx(mass, abs=1e-6)


def test_total_cable():
    export = ElectricalDesign(config)
    export.run()

    length = 0.02 + 3 + 30
    length += length * 0.01
    mass = length * export.cable.linear_density
    assert export.total_mass == pytest.approx(mass * 9, abs=1e-10)
    assert export.total_length == pytest.approx(length * 9, abs=1e-10)


def test_total_cable_cost():
    export = ElectricalDesign(config)
    export.run()

    assert export.total_cable_cost == 135068310.0


def test_cables_property():
    export = ElectricalDesign(config)
    export.run()

    assert (
        export.sections_cables == export.cable.name
    ).sum() == export.num_cables


def test_cable_lengths_property():
    export = ElectricalDesign(config)
    export.run()

    cable_name = export.cable.name
    assert (
        export.cable_lengths_by_type[cable_name] == export.length
    ).sum() == export.num_cables


def test_total_cable_len_property():
    export = ElectricalDesign(config)
    export.run()

    cable_name = export.cable.name
    assert export.total_cable_length_by_type[cable_name] == pytest.approx(
        export.total_length,
        abs=1e-10,
    )


def test_design_result():
    export = ElectricalDesign(config)
    export.run()

    _ = export.cable.name
    cables = export.design_result["export_system"]["cable"]

    assert cables["sections"] == [export.length]
    assert cables["number"] == 9
    assert cables["linear_density"] == export.cable.linear_density


def test_floating_length_calculations():
    base = deepcopy(config)
    base["site"]["depth"] = 250
    base["export_system_design"]["touchdown_distance"] = 0

    sim = ElectricalDesign(base)
    sim.run()

    base_length = sim.total_length

    with_cat = deepcopy(config)
    with_cat["site"]["depth"] = 250

    new = ElectricalDesign(with_cat)
    new.run()

    assert new.total_length < base_length


def test_HVDC_cable():
    base = deepcopy(config)
    base["export_system_design"] = {"cables": "HVDC_2000mm_320kV"}

    sim = ElectricalDesign(base)
    sim.run()

    assert sim.num_cables % 2 == 0

    base = deepcopy(config)
    base["export_system_design"] = {"cables": "HVDC_2500mm_525kV"}

    sim = ElectricalDesign(base)
    sim.run()

    assert sim.num_cables % 2 == 0


def test_num_crossing():
    base_sim = ElectricalDesign(config)
    base_sim.run()

    cross = deepcopy(config)
    cross["export_system_design"]["cable_crossings"] = {"crossing_number": 2}

    cross_sim = ElectricalDesign(cross)
    cross_sim.run()

    assert cross_sim.crossing_cost != base_sim.crossing_cost


def test_cost_crossing():
    base_sim = ElectricalDesign(config)
    base_sim.run()

    cross = deepcopy(config)
    cross["export_system_design"]["cable_crossings"] = {
        "crossing_number": 1,
        "crossing_unit_cost": 100000,
    }

    cross_sim = ElectricalDesign(cross)
    cross_sim.run()

    assert cross_sim.crossing_cost != base_sim.crossing_cost


def test_deprecated_landfall():

    base = deepcopy(config)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        sim = ElectricalDesign(base)
        sim.run()

    deprecated = deepcopy(base)
    deprecated["landfall"] = {"interconnection_distance": 4}

    with pytest.deprecated_call():
        sim = ElectricalDesign(deprecated)
        sim.run()



"""Tests for the `MooringSystemDesign` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import pytest

from wisdem.orbit.phases.design import MooringSystemDesign

base = {
    "site": {"depth": 200},
    "turbine": {"turbine_rating": 6},
    "plant": {"num_turbines": 50},
    "mooring_system_design": {},
}


@pytest.mark.parametrize("depth", range(10, 1000, 100))
def test_depth_sweep(depth):

    config = deepcopy(base)
    config["site"]["depth"] = depth

    moor = MooringSystemDesign(config)
    moor.run()

    assert moor.design_result
    assert moor.total_cost


@pytest.mark.parametrize("rating", range(3, 15, 1))
def test_rating_sweep(rating):

    config = deepcopy(base)
    config["turbine"]["turbine_rating"] = rating

    moor = MooringSystemDesign(config)
    moor.run()

    assert moor.design_result
    assert moor.total_cost


def test_mooring_system_defaults():

    moor_base = MooringSystemDesign(base)
    moor_base.run()

    base_cost = moor_base.detailed_output["system_cost"]

    config_defs = deepcopy(base)
    config_defs["mooring_system_design"] = {}
    config_defs["mooring_system_design"]["mooring_type"] = "Catenary"
    config_defs["mooring_system_design"]["anchor_type"] = "Suction Pile"

    moor_defs = MooringSystemDesign(config_defs)
    moor_defs.run()

    assert moor_defs.detailed_output["system_cost"] == base_cost


def test_catenary_mooring_system_kwargs():

    test_kwargs = {
        "num_lines": 6,
        "anchor_type": "Drag Embedment",
        "mooring_line_cost_rate": 2500,
    }

    moor = MooringSystemDesign(base)
    moor.run()

    base_cost = moor.detailed_output["system_cost"]

    assert base_cost == pytest.approx(76173891, abs=1e0)

    for k, v in test_kwargs.items():
        config = deepcopy(base)
        config["mooring_system_design"] = {}
        config["mooring_system_design"][k] = v

        moor = MooringSystemDesign(config)
        moor.run()

        assert moor.detailed_output["system_cost"] != base_cost


def test_semitaut_mooring_system_kwargs():

    semi_base = deepcopy(base)
    semi_base["mooring_system_design"]["mooring_type"] = "SemiTaut"

    test_kwargs = {
        "num_lines": 6,
        "anchor_type": "Drag Embedment",
        "chain_density": 10000,
        "rope_density": 1000,
    }

    moor = MooringSystemDesign(semi_base)
    moor.run()

    base_cost = moor.detailed_output["system_cost"]

    assert base_cost == pytest.approx(102227311, abs=1e0)

    for k, v in test_kwargs.items():
        config = deepcopy(semi_base)
        config["mooring_system_design"] = {}
        config["mooring_system_design"][k] = v

        moor = MooringSystemDesign(config)
        moor.run()

        assert moor.detailed_output["system_cost"] != base_cost


def test_tlp_mooring_system_kwargs():

    tlp_base = deepcopy(base)
    tlp_base["mooring_system_design"]["mooring_type"] = "TLP"

    test_kwargs = {
        "num_lines": 6,
        "anchor_type": "Drag Embedment",
        "mooring_line_cost_rate": 2500,
        "draft_depth": 10,
    }

    moor = MooringSystemDesign(tlp_base)
    moor.run()

    base_cost = moor.detailed_output["system_cost"]

    assert base_cost == pytest.approx(57633231, abs=1e0)

    for k, v in test_kwargs.items():
        config = deepcopy(tlp_base)
        config["mooring_system_design"] = {}
        config["mooring_system_design"][k] = v

        moor = MooringSystemDesign(config)
        moor.run()

        assert moor.detailed_output["system_cost"] != base_cost


def test_drag_embedment_fixed_length():

    moor = MooringSystemDesign(base)
    moor.run()

    baseline = moor.line_length

    default = deepcopy(base)
    default["mooring_system_design"] = {"anchor_type": "Drag Embedment"}

    moor = MooringSystemDesign(default)
    moor.run()

    with_default = moor.line_length
    assert with_default > baseline

    custom = deepcopy(base)
    custom["mooring_system_design"] = {
        "anchor_type": "Drag Embedment",
        "drag_embedment_fixed_length": 1000,
    }

    moor = MooringSystemDesign(custom)
    moor.run()

    assert moor.line_length > with_default
    assert moor.line_length > baseline


def test_custom_num_lines():

    config = deepcopy(base)
    config["mooring_system_design"] = {"num_lines": 5}

    moor = MooringSystemDesign(config)
    moor.run()

    assert moor.design_result["mooring_system"]["num_lines"] == 5



"""Provides the `ElectricalDesign` class."""

__author__ = ["Sophie Bredenkamp"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = ""
__email__ = []

from warnings import warn

import numpy as np

from wisdem.orbit.phases.design._cables import CableSystem

"""
[1] Maness et al. 2017, NREL Offshore Balance-of-System Model.
https://www.nrel.gov/docs/fy17osti/66874.pdf
"""


class ElectricalDesign(CableSystem):
    """
    Design phase for export cabling and offshore substation systems.

    Attributes
    ----------
    num_cables : int
        Total number of cables required for transmitting power.
    length : float
        Length of a single cable connecting the OSS to the interconnection
        in km.
    mass : float
        Mass of `length` in tonnes.
    cable : `Cable`
        Instance of `ORBIT.phases.design.Cable`. An export system will
        only require a single type of cable.
    total_length : float
        Total length of cable required to trasmit power.
    total_mass : float
        Total mass of cable required to transmit power.
    sections_cables : np.ndarray, shape: (`num_cables, )
        An array of `cable`.
    sections_lengths : np.ndarray, shape: (`num_cables, )
        An array of `length`.

    """

    #:
    expected_config = {
        "site": {"distance_to_landfall": "km", "depth": "m"},
        "landfall": {"interconnection_distance": "km (optional)"},
        "plant": {"capacity": "MW"},
        "export_system_design": {
            "cables": "str",
            "num_redundant": "int (optional)",
            "touchdown_distance": "m (optional, default: 0)",
            "percent_added_length": "float (optional)",
            "interconnection_distance": "km (optional)",
            "cable_crossings": {
                "crossing_number": "int (optional)",
                "crossing_unit_cost": "float (optional)",
            },
        },
        "substation_design": {
            "substation_capacity": "MW (optional)",
            "num_substations": "int (optional)",
            "mpt_unit_cost": "USD/cable (optional)",
            "topside_design_cost": "USD (optional)",
            "shunt_unit_cost": "USD/cable (optional)",
            "switchgear_cost": "USD (optional)",
            "dc_breaker_cost": "USD (optional)",
            "backup_gen_cost": "USD (optional)",
            "workspace_cost": "USD (optional)",
            "other_ancillary_cost": "USD (optional)",
            "converter_cost": "USD (optional)",
            "onshore_converter_cost": "USD (optional)",
            "topside_assembly_factor": "float (optional)",
            "oss_substructure_type": "str (optional, default: Monopile)",
            "oss_substructure_cost_rate": "USD/t (optional)",
            "oss_pile_cost_rate": "USD/t (optional)",
        },
        "onshore_substation_design": {
            "shunt_unit_cost": "USD/cable (optional)",
            "onshore_converter_cost": "USD (optional)",
        },
    }

    output_config = {
        "num_substations": "int",
        "offshore_substation_topside": "dict",
        "offshore_substation_substructure": "dict",
        "export_system": {
            "system_cost": "USD",
            "cable": {
                "linear_density": "t/km",
                "sections": [("length, km", "speed, km/h (optional)")],
                "number": "int (optional)",
                "diameter": "int",
                "cable_type": "str",
            },
        },
        "offshore_substation": "dict, (optional)",
    }

    def __init__(self, config, **kwargs):
        """Creates an instance of ElectricalDesign."""

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        # CABLES
        super().__init__(config, "export", **kwargs)

        for name in self.expected_config["site"]:
            setattr(self, "".join(("_", name)), config["site"][name])

        self._depth = config["site"]["depth"]
        self._distance_to_landfall = config["site"]["distance_to_landfall"]
        self._plant_capacity = self.config["plant"]["capacity"]
        self._get_touchdown_distance()

        self._design = self.config["export_system_design"]

        _landfall = self.config.get("landfall", {})
        if _landfall:
            warn(
                "landfall dictionary will be deprecated and moved"
                " into [export_system_design][landfall].",
                DeprecationWarning,
                stacklevel=2,
            )

        else:
            _landfall = self._design.get("landfall", {})

        self._distance_to_interconnection = _landfall.get(
            "interconnection_distance", 3
        )

        self._oss_design = self.config.get("substation_design", {})

        self.substructure_type = self._oss_design.get(
            "oss_substructure_type", "Monopile"
        ).title()

        self._outputs = {}

    def run(self):
        """Main run function."""

        # CABLES
        self._initialize_cables()
        self.cable = self.cables[[*self.cables][0]]
        self.compute_number_cables()
        self.compute_cable_length()
        self.compute_cable_mass()
        self.compute_total_cable()
        self.calc_crossing_cost()

        self._outputs["export_system"] = {
            "landfall": {
                "interconnection_distance": (self._distance_to_interconnection)
            },
            "system_cost": self.total_cable_cost,
        }

        for cable in self.cables.values():
            self._outputs["export_system"]["cable"] = {
                "linear_density": cable.linear_density,
                "sections": [self.length],
                "number": self.num_cables,
                "cable_power": cable.cable_power,
                "cable_type": cable.cable_type,
            }

        # SUBSTATION
        self.calc_num_substations()
        self.calc_substructure_length()
        self.calc_substructure_deck_space()
        self.calc_topside_deck_space()

        self.calc_mpt_cost()
        self.calc_topside_mass_and_cost()
        self.calc_shunt_reactor_cost()
        self.calc_switchgear_costs()
        self.calc_ancillary_system_cost()
        self.calc_assembly_cost()
        self.calc_substructure_mass_and_cost()
        self.calc_converter_cost()
        self.calc_dc_breaker_cost()
        self.calc_onshore_cost()

        self._outputs["offshore_substation"] = {
            "substation_mpt_cost": self.mpt_cost,
            "substation_shunt_cost": self.shunt_reactor_cost,
            "substation_switchgear_cost": self.switchgear_cost,
            "substation_converter_cost": self.converter_cost,
            "substation_breaker_cost": self.dc_breaker_cost,
            "substation_ancillary_cost": self.ancillary_system_costs,
            "substation_land_assembly_cost": self.land_assembly_cost,
        }

        self._outputs["offshore_substation_substructure"] = {
            "type": self.substructure_type,
            "deck_space": self.substructure_deck_space,
            "mass": self.substructure_mass,
            "length": self.substructure_length,
            "unit_cost": self.substructure_cost,
        }

        # TODO: cheap fix for topside unit_cost bug #168
        self._outputs["offshore_substation_topside"] = {
            "deck_space": self.topside_deck_space,
            "mass": self.topside_mass,
            "unit_cost": self.substation_cost + self.topside_cost,
        }

        self._outputs["num_substations"] = self.num_substations
        self._outputs["total_substation_cost"] = self.total_substation_cost

    @property
    def detailed_output(self):
        """Returns export system design outputs."""

        _output = {
            **self.design_result,
            "export_system_total_mass": self.total_mass,
            "export_system_total_length": self.total_length,
            "export_system_total_cost": self.total_cable_cost,
            "export_system_cable_power": self.cable.cable_power,
            "num_substations": self.num_substations,
            "substation_mpt_rating": self.mpt_rating,
            "substation_topside_mass": self.topside_mass,
            "substation_topside_cost": self.topside_cost,
            "substation_substructure_mass": self.substructure_mass,
            "substation_substructure_cost": self.substructure_cost,
            "total_substation_cost": self.total_substation_cost,
            "substation_mpt_cost": self.mpt_cost,
            "substation_shunt_cost": self.shunt_reactor_cost,
            "substation_switchgear_cost": self.switchgear_cost,
            "substation_converter_cost": self.converter_cost,
            "substation_breaker_cost": self.dc_breaker_cost,
            "substation_ancillary_cost": self.ancillary_system_costs,
            "substation_land_assembly_cost": self.land_assembly_cost,
            "onshore_shunt_cost": self.onshore_shunt_reactor_cost,
            "onshore_converter_cost": self.onshore_converter_cost,
            "onshore_switchgear_cost": self.onshore_switchgear_cost,
            "onshore_construction_cost": self.onshore_construction,
            "onshore_compensation_cost": self.onshore_compensation_cost,
            "onshore_mpt_cost": self.mpt_cost,
        }

        return _output

    @property
    def design_result(self):
        """Returns the results of self.run()."""
        return self._outputs

        # CABLES

    @property
    def total_cable_cost(self):
        """Returns total export system cable cost."""

        return sum(self.cost_by_type.values()) + self.crossing_cost

    def compute_number_cables(self):
        """
        Calculate the total number of required and redundant cables to
        transmit power to the onshore interconnection.

        """

        num_required = np.ceil(self._plant_capacity / self.cable.cable_power)
        num_redundant = self._design.get("num_redundant", 0)

        if "HVDC" in self.cable.cable_type:
            num_required *= 2
            num_redundant *= 2

        self.num_cables = int(num_required + num_redundant)

    def compute_cable_length(self):
        """Calculates the total distance an export cable must travel."""

        added_length = 1.0 + self._design.get("percent_added_length", 0.0)
        self.length = round(
            (
                self.free_cable_length
                + (self._distance_to_landfall - self.touchdown / 1000)
                + self._distance_to_interconnection
            )
            * added_length,
            10,
        )

    def compute_cable_mass(self):
        """Calculates the total mass of a single length of export cable."""

        self.mass = round(self.length * self.cable.linear_density, 10)

    def compute_total_cable(self):
        """
        Calculates the total length and mass of cables required to fully
        connect the OSS to the interconnection point.
        """

        self.total_length = round(self.num_cables * self.length, 10)
        self.total_mass = round(self.num_cables * self.mass, 10)

    @property
    def sections_cable_lengths(self):
        """
        Creates an array of section lengths to work with ``CableSystem``.

        Returns
        -------
        np.ndarray
            Array of `length` with shape (``num_cables``, ).
        """
        return np.full(self.num_cables, self.length)

    @property
    def sections_cables(self):
        """
        Creates an array of cable names to work with ``CableSystem``.

        Returns
        -------
        np.ndarray
            Array of ``cable.name`` with shape (``num_cables``, ).
        """

        return np.full(self.num_cables, self.cable.name)

    def calc_crossing_cost(self):
        """Compute cable crossing costs."""
        _crossing_design = self._design.get("cable_crossings", {})

        _key = "crossing_unit_cost"
        crossing_cost = _crossing_design.get(
            _key,
            self.get_default_cost(
                "export_system_design", "cable_crossings", subkey=_key
            ),
        )

        self.crossing_cost = crossing_cost * _crossing_design.get(
            "crossing_number", 0
        )

        """SUBSTATION"""

    @property
    def total_substation_cost(self):
        """Returns the total substation cost."""

        return (
            self.topside_cost + self.substructure_cost + self.substation_cost
        )

    def calc_num_substations(self):
        """Computes number of substations based on HVDC or HVAC
        export cables.
        """

        # HVAC substation capacity
        _substation_capacity = self._oss_design.get(
            "substation_capacity", 1200
        )  # MW

        if "HVDC" in self.cable.cable_type:
            self.num_substations = self._oss_design.get(
                "num_substations", int(self.num_cables / 2)
            )
        else:
            self.num_substations = self._oss_design.get(
                "num_substations",
                int(np.ceil(self._plant_capacity / _substation_capacity)),
            )

    @property
    def substation_cost(self):
        """Returns total procuremet cost of the topside."""

        return (
            self.mpt_cost
            + self.shunt_reactor_cost
            + self.switchgear_cost
            + self.converter_cost
            + self.dc_breaker_cost
            + self.ancillary_system_costs
            + self.land_assembly_cost
        ) / self.num_substations

    def calc_mpt_cost(self):
        """Computes HVAC main power transformer (MPT). MPT cost is 0 for
        HVDC.
        """

        _key = "mpt_unit_cost"
        _mpt_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        self.num_mpt = self.num_cables

        self.mpt_cost = (
            0 if "HVDC" in self.cable.cable_type else self.num_mpt * _mpt_cost
        )

        self.mpt_rating = (
            round((self._plant_capacity * 1.15 / self.num_mpt) / 10.0) * 10.0
        )

    def calc_shunt_reactor_cost(self):
        """Computes HVAC shunt reactor cost. Shunt reactor cost is 0 for
        HVDC.
        """

        touchdown = self.config["site"]["distance_to_landfall"]

        _key = "shunt_unit_cost"

        shunt_unit_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        if "HVDC" in self.cable.cable_type:
            self.compensation = 0
        else:
            for cable in self.cables.values():
                self.compensation = touchdown * cable.compensation_factor  # MW

        self.shunt_reactor_cost = (
            self.compensation * shunt_unit_cost * self.num_cables
        )

    def calc_switchgear_costs(self):
        """Computes HVAC switchgear cost. Switchgear cost is 0 for HVDC."""

        _key = "switchgear_cost"
        switchgear_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        self.num_switchgear = (
            0 if "HVDC" in self.cable.cable_type else self.num_cables
        )

        self.switchgear_cost = self.num_switchgear * switchgear_cost

    def calc_dc_breaker_cost(self):
        """Computes HVDC circuit breaker cost. Breaker cost is 0 for HVAC."""

        _key = "dc_breaker_cost"
        dc_breaker_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        num_dc_breakers = (
            self.num_cables if "HVDC" in self.cable.cable_type else 0
        )

        self.dc_breaker_cost = num_dc_breakers * dc_breaker_cost

    def calc_ancillary_system_cost(self):
        """Calculates cost of ancillary systems."""

        _key = "backup_gen_cost"
        backup_gen_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        _key = "workspace_cost"
        workspace_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        _key = "other_ancillary_cost"
        other_ancillary_cost = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        self.ancillary_system_costs = (
            backup_gen_cost + workspace_cost + other_ancillary_cost
        ) * self.num_substations

    def calc_assembly_cost(self):
        """Calculates the cost of assembly on land."""

        _key = "topside_assembly_factor"
        topside_assembly_factor = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        if topside_assembly_factor > 1.0:
            topside_assembly_factor /= 100

        self.land_assembly_cost = (
            self.switchgear_cost
            + self.shunt_reactor_cost
            + self.ancillary_system_costs
        ) * topside_assembly_factor

    def calc_converter_cost(self):
        """Computes converter cost."""

        _key = "converter_cost"
        converter_cost = self._oss_design.get(
            _key,
            self.get_default_cost(
                "substation_design", _key, subkey=self.cable.cable_type
            ),
        )

        self.converter_cost = converter_cost

    def calc_substructure_mass_and_cost(self):
        """
        Calculates the mass and associated cost of the substation substructure
        based on equations 81-84 [1].
        """

        _key = "oss_substructure_cost_rate"
        oss_substructure_cost_rate = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        _key = "oss_pile_cost_rate"
        oss_pile_cost_rate = self._oss_design.get(
            _key, self.get_default_cost("substation_design", _key)
        )

        # Substructure mass components
        # TODO: Determine a better method to calculate substructure mass
        #       for different substructure types
        substructure_mass = 0.4 * self.topside_mass

        substructure_pile_mass = (
            0
            if self.substructure_type == "Floating"
            else 8 * substructure_mass**0.5574
        )

        self.substructure_cost = (
            substructure_mass * oss_substructure_cost_rate
            + substructure_pile_mass * oss_pile_cost_rate
        )

        self.substructure_mass = substructure_mass + substructure_pile_mass

    def calc_substructure_length(self):
        """Calculates substructure length as the site depth + 10m."""

        if self.substructure_type == "Floating":
            self.substructure_length = 0

        else:
            self.substructure_length = self.config["site"]["depth"] + 10

    def calc_substructure_deck_space(self):
        """
        Calculates required deck space for the substation substructure.

        Coming soon!
        """

        self.substructure_deck_space = 1

    def calc_topside_deck_space(self):
        """
        Calculates required deck space for the substation topside.

        Coming soon!
        """

        self.topside_deck_space = 1

    def calc_topside_mass_and_cost(self):
        """Calculates the mass and cost of the substation topsides."""

        self.topside_mass = (
            3.85 * (self.mpt_rating * self.num_mpt) / self.num_substations
            + 285
        )

        _key = "topside_design_cost"
        topside_design_cost = self._oss_design.get(
            _key,
            self.get_default_cost(
                "substation_design", _key, subkey=self.cable.cable_type
            ),
        )

        self.topside_cost = topside_design_cost

    def calc_onshore_cost(self):
        """Minimum Cost of Onshore Substation Connection."""

        _design = self.config.get("onshore_substation_design", {})

        _key = "onshore_converter_cost"
        _converter_cost = _design.get(
            _key,
            self.get_default_cost(
                "onshore_substation_design", _key, subkey=self.cable.cable_type
            ),
        )

        self.onshore_converter_cost = self.num_substations * _converter_cost

        _key = "switchgear_cost"
        _switchgear_cost = _design.get(
            _key, self.get_default_cost("onshore_substation_design", _key)
        )

        self.onshore_switchgear_cost = self.num_switchgear * _switchgear_cost

        _key = "onshore_construction_rate"
        _construction_rate = _design.get(
            _key,
            self.get_default_cost(
                "onshore_substation_design", _key, subkey=self.cable.cable_type
            ),
        )

        self.onshore_construction = self.num_substations * _construction_rate

        _key = "shunt_unit_cost"
        _shunt_unit_cost = _design.get(
            _key, self.get_default_cost("onshore_substation_design", _key)
        )

        self.onshore_shunt_reactor_cost = (
            self.compensation * self.num_cables * _shunt_unit_cost
        )

        _key = "compensation_rate"
        _compensation_rate = _design.get(
            _key,
            self.get_default_cost(
                "onshore_substation_design", _key, subkey=self.cable.cable_type
            ),
        )

        self.onshore_compensation_cost = (
            self.num_cables * _compensation_rate
            + self.onshore_shunt_reactor_cost
        )

        self.onshore_cost = (
            self.onshore_converter_cost
            + self.onshore_switchgear_cost
            + self.onshore_construction
            + self.onshore_compensation_cost
            + self.mpt_cost
        )

        self._outputs["export_system"][
            "onshore_substation_costs"
        ] = self.onshore_cost



"""Provides the configuration loading and saving methods."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from pathlib import Path

import yaml
from yaml import Dumper

from wisdem.orbit.core import loader


def load_config(filepath):
    """
    Load an ORBIT config at `filepath`.

    Parameters
    ----------
    filepath : str
        Path to yaml config file.
    """

    with Path(filepath).open() as f:
        data = yaml.load(f, Loader=loader)

    return data


def save_config(config, filepath, overwrite=False):
    """
    Save an ORBIT `config` to `filepath`.

    Parameters
    ----------
    config : dict
        ORBIT configuration.
    filepath : str
        Location to save config.
    overwrite : bool (optional)
        Overwrite file if it already exists. Default: False.
    """

    filepath = Path(filepath).resolve()
    dirs = filepath.parent
    if not dirs.exists():
        dirs.mkdir(parents=True)

    if overwrite is False:
        if filepath.exists():
            raise FileExistsError(f"File already exists at '{filepath}'.")

    with filepath.open("w") as f:
        yaml.dump(config, f, Dumper=Dumper, default_flow_style=False)



import os

import numpy as np
import openmdao.api as om
from scipy.interpolate import PchipInterpolator


class ParametrizeBladeAero(om.ExplicitComponent):
    # Openmdao component to parameterize distributed quantities for the outer shape of the wind turbine rotor blades
    def initialize(self):
        self.options.declare("rotorse_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["rotorse_options"]
        self.opt_options = self.options["opt_options"]
        n_span = rotorse_options["n_span"]
        self.n_opt_twist = n_opt_twist = self.opt_options["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"]
        self.n_opt_chord = n_opt_chord = self.opt_options["design_variables"]["blade"]["aero_shape"]["chord"]["n_opt"]

        # Inputs
        self.add_input(
            "s",
            val=np.zeros(n_span),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis (0-blade root, 1-blade tip)",
        )
        # # Blade twist
        # self.add_input(
        #     "twist_original",
        #     val=np.zeros(n_span),
        #     units="rad",
        #     desc="1D array of the twist values defined along blade span. The twist is the one defined in the yaml.",
        # )
        self.add_input(
            "s_opt_twist",
            val=np.zeros(n_opt_twist),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade twist angle",
        )
        self.add_input(
            "twist_opt",
            val=np.ones(n_opt_twist),
            units="rad",
            desc="1D array of the twist angle being optimized at the n_opt locations.",
        )
        # Blade chord
        # self.add_input(
        #     "chord_original",
        #     val=np.zeros(n_span),
        #     units="m",
        #     desc="1D array of the chord values defined along blade span. The chord is the one defined in the yaml.",
        # )
        self.add_input(
            "s_opt_chord",
            val=np.zeros(n_opt_chord),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis to optimize the blade chord",
        )
        self.add_input(
            "chord_opt",
            val=np.ones(n_opt_chord),
            units="m",
            desc="1D array of the chord being optimized at the n_opt locations.",
        )

        # Outputs
        self.add_output(
            "twist_param",
            val=np.zeros(n_span),
            units="rad",
            desc="1D array of the twist values defined along blade span. The twist is the result of the parameterization.",
        )
        self.add_output(
            "chord_param",
            val=np.zeros(n_span),
            units="m",
            desc="1D array of the chord values defined along blade span. The chord is the result of the parameterization.",
        )
        self.add_output(
            "max_chord_constr",
            val=np.zeros(n_opt_chord),
            desc="1D array of the ratio between chord values and maximum chord along blade span.",
        )
        self.add_output(
            "slope_chord_constr",
            val=np.zeros(n_opt_chord-1),
            desc="1D array of the difference between one chord point and the other. If larger than 0, chord is growing along span. It can be used as constraint to achieve monotically decreasing chord",
        )

    def compute(self, inputs, outputs):
        spline = PchipInterpolator
        twist_spline = spline(inputs["s_opt_twist"], inputs["twist_opt"])
        outputs["twist_param"] = twist_spline(inputs["s"])
        chord_spline = spline(inputs["s_opt_chord"], inputs["chord_opt"])
        outputs["chord_param"] = chord_spline(inputs["s"])
        chord_opt = spline(inputs["s"], outputs["chord_param"])
        max_chord = self.opt_options["constraints"]["blade"]["chord"]["max"]
        outputs["max_chord_constr"] = chord_opt(inputs["s_opt_chord"]) / max_chord
        # Define constraint to enforce monothonically decreasing blade chord
        outputs["slope_chord_constr"] = np.diff(inputs["chord_opt"])
        # Pick the points before max chord or before 40% span and deactivate the constraint
        id_max_chord = np.argmax(inputs["chord_opt"])
        if inputs["s_opt_chord"][id_max_chord] > 0.4:
            id_start_constraint = id_max_chord
        else:
            id_start_constraint = np.argmin(abs(inputs["s_opt_chord"]-0.4))
        outputs["slope_chord_constr"][:id_start_constraint] = -1 


class ParametrizeBladeStruct(om.ExplicitComponent):
    # Openmdao component to parameterize distributed quantities for the structural design of the wind turbine rotor blades
    def initialize(self):
        self.options.declare("rotorse_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["rotorse_options"]
        self.opt_options = opt_options = self.options["opt_options"]
        self.n_span = n_span = rotorse_options["n_span"]
        self.n_layers = n_layers = rotorse_options["n_layers"]
        # Inputs
        self.add_input(
            "s",
            val=np.zeros(n_span),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis (0-blade root, 1-blade tip)",
        )
        self.add_input(
            "layer_thickness_original",
            val=np.zeros((n_layers, n_span)),
            units="m",
            desc="2D array of the thickness of the layers of the blade structure. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )

        for i in range(n_layers):
            self.add_input(
                "s_opt_layer_%d"%i,
                val=np.ones(opt_options["design_variables"]["blade"]["n_opt_struct"][i]),
            )
            self.add_input(
                "layer_%d_opt"%i,
                units="m",
                val=np.ones(opt_options["design_variables"]["blade"]["n_opt_struct"][i]),
            )

        # Outputs
        self.add_output(
            "layer_thickness_param",
            val=np.zeros((n_layers, n_span)),
            units="m",
            desc="2D array of the thickness of the layers of the blade structure after the parametrization. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )

    def compute(self, inputs, outputs):

        for i in range(self.n_layers):
            opt_m_interp = PchipInterpolator(inputs["s_opt_layer_%d"%i], inputs["layer_%d_opt"%i])(inputs["s"])
            outputs["layer_thickness_param"][i, :] = opt_m_interp

class ComputeReynolds(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("n_span")

    def setup(self):
        n_span = self.options["n_span"]

        self.add_input("rho", val=0.0, units="kg/m**3")
        self.add_input("mu", val=1.81e-5, units="kg/(m*s)", desc="Dynamic viscosity of air")
        self.add_input("chord", val=np.zeros((n_span)), units="m")
        self.add_input("r_blade", val=np.zeros(n_span), units="m",
            desc="1D array of the dimensional spanwise grid defined along the rotor (hub radius to blade tip projected on the plane)",
        )
        self.add_input("rotor_radius", val=0.0, units="m",
            desc="Scalar of the rotor radius, defined ignoring prebend and sweep curvatures, and cone and uptilt angles.",
        )
        self.add_input("maxOmega", val=0.0, units="rad/s", desc="Maximum allowed rotor speed.")
        self.add_input("max_TS", val=0.0, units="m/s", desc="Maximum allowed blade tip speed.")
        self.add_input("V_out", val=0.0, units="m/s", desc="Cut out wind speed. This is the wind speed where region III ends.")

        self.add_output("Re", val=np.zeros((n_span)), ref=1.0e6)
        
    def compute(self, inputs, outputs):
        # Note that we used to use ccblade outputs of local wind speed at the rated condition
        # This is more accurate, of course, but creates an implicit feedback loop in the code
        # This way gets an order-of-magnitude estimate for Reynolds number, which is really all that is needed
        max_local_TS = inputs["max_TS"] / inputs["rotor_radius"] * inputs["r_blade"]
        if np.all(max_local_TS == 0.0):
            max_local_TS = inputs["maxOmega"] * inputs["r_blade"]

        max_local_V = np.sqrt(inputs["V_out"]**2 + max_local_TS**2)
        outputs["Re"] = np.nan_to_num(
            inputs["rho"] * max_local_V * inputs["chord"] / inputs["mu"]
        )



"""`MooringSystemDesign` and related functionality."""

__author__ = "Jake Nunemaker, Becca Fuchs"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Nicholas Riccobono"
__email__ = (
    "jake.nunemaker@nrel.gov, rebecca.fuchs@nrel.gov,"
    "nicholas.riccobono@nrel.gov"
)

from math import sqrt

from scipy.interpolate import interp1d

from wisdem.orbit.phases.design import DesignPhase

"""
[1] Maness et al. 2017, NREL Offshore Balance-of-System Model.
https://www.nrel.gov/docs/fy17osti/66874.pdf

[2] Cooperman et al. (2022), Assessment of Offshore Wind Energy Leasing Areas
for Humboldt and Morry Bay. https://www.nrel.gov/docs/fy22osti/82341.pdf
"""


class MooringSystemDesign(DesignPhase):
    """Mooring System and Anchor Design."""

    expected_config = {
        "site": {"depth": "float"},
        "turbine": {"turbine_rating": "int | float"},
        "plant": {"num_turbines": "int"},
        "mooring_system_design": {
            "num_lines": "int | float (optional, default: 4)",
            "anchor_type": "str (optional, default: 'Suction Pile')",
            "mooring_type": "str (optional, default: 'Catenary')",
            "mooring_line_cost_rate": "int | float (optional)",
            "drag_embedment_fixed_length": "int (optional, default: 500m)",
            "draft_depth": "int (optional, default: 20m)",
            "chain_density": "int | float (optional, default: 19900 kg/m**3)",
            "rope_density": "int | float (optional, default: 797.8 kg/m**3)",
        },
    }

    output_config = {
        "mooring_system": {
            "num_lines": "int",
            "line_diam": "m, float",
            "line_mass": "t",
            "line_cost": "USD",
            "line_length": "m",
            "mooring_type": "str",
            "anchor_mass": "t",
            "anchor_type": "str",
            "anchor_cost": "USD",
            "system_cost": "USD",
        }
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of MooringSystemDesign.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self.num_turbines = self.config["plant"]["num_turbines"]
        self.depth = self.config["site"]["depth"]

        self._design = self.config.get("mooring_system_design", {})
        self.num_lines = self._design.get("num_lines", 4)
        self.anchor_type = self._design.get(
            "anchor_type", "Suction Pile"
        ).title()
        self.mooring_type = self._design.get(
            "mooring_type", "Catenary"
        ).title()

        # Semi-Taut mooring system design parameters based on depth [2].
        self._semitaut_params = {
            "depths": [500.0, 750.0, 1000.0, 1250.0, 1500.0],
            "rope_lengths": [478.41, 830.34, 1229.98, 1183.93, 1079.62],
            "rope_diameter": 0.2,
            "chain_lengths": [917.11, 800.36, 609.07, 896.42, 1280.57],
            "chain_diameters": [0.13, 0.17, 0.22, 0.22, 0.22],
            "anchor_costs": [112766.0, 125511.0, 148703.0, 204988.0, 246655.0],
            "total_line_costs": [
                826598.0,
                1221471.0,
                1682208.0,
                2380035.0,
                3229700.0,
            ],
        }

        self._outputs = {}

    def run(self):
        """Runs the design model."""

        self.determine_mooring_line()
        self.calculate_breaking_load()
        self.calculate_line_length_mass()
        self.calculate_anchor_mass_cost()

        self._outputs["mooring_system"] = {**self.design_result}

    def determine_mooring_line(self):
        """
        Returns the diameter of the mooring lines based on the turbine rating.

        TODO: Add TLP option and consider merging SemiTaut interp here
        """

        tr = self.config["turbine"]["turbine_rating"]
        fit = -0.0004 * (tr**2) + 0.0132 * tr + 0.0536

        _key = "mooring_line_cost_rate"

        mooring_line_cost_rate = self._design.get(
            _key,
            self.get_default_cost(
                "mooring_system_design",
                _key,
            ),
        )
        if isinstance(mooring_line_cost_rate, (int, float)):
            mooring_line_cost_rate = [mooring_line_cost_rate] * 3

        if fit <= 0.09:
            self.line_diam = 0.09
            self.line_mass_per_m = 0.161
            self.line_cost_rate = mooring_line_cost_rate[0]

        elif fit <= 0.12:
            self.line_diam = 0.12
            self.line_mass_per_m = 0.288
            self.line_cost_rate = mooring_line_cost_rate[1]

        else:
            self.line_diam = 0.15
            self.line_mass_per_m = 0.450
            self.line_cost_rate = mooring_line_cost_rate[2]

    def calculate_breaking_load(self):
        """Returns the mooring line breaking load."""

        self.breaking_load = (
            419449 * (self.line_diam**2) + 93415 * self.line_diam - 3577.9
        )

    def calculate_line_length_mass(self):
        """
        Returns the mooring line length and mass.

        SemiTaut model based on:
        https://github.com/NREL/MoorPy/blob/dev/moorpy/MoorProps_default.yaml

        TODO: Improve TLP line length and mass

        """

        # Add extra fixed line length for drag embedments
        if self.anchor_type == "Drag Embedment":
            fixed = self._design.get("drag_embedment_fixed_length", 500)

        else:
            fixed = 0

        draft = self._design.get("draft_depth", 20)

        if self.mooring_type == "Semitaut":

            # Interpolation of rope and chain length at project depth
            self.chain_length = interp1d(
                self._semitaut_params["depths"],
                self._semitaut_params["chain_lengths"],
                fill_value="extrapolate",
            )(self.depth).item()
            self.rope_length = interp1d(
                self._semitaut_params["depths"],
                self._semitaut_params["rope_lengths"],
                fill_value="extrapolate",
            )(self.depth).item()

            # Rope and interpolated chain diameter at project depth
            rope_diameter = self._semitaut_params["rope_diameter"]
            chain_diameter = interp1d(
                self._semitaut_params["depths"],
                self._semitaut_params["chain_diameters"],
                fill_value="extrapolate",
            )(self.depth).item()

            fixed = self._design.get("drag_embedment_fixed_length", 0)
            self.line_length = self.rope_length + self.chain_length + fixed

            # line characteristics based on MoorPy defaults,
            chain_mass_per_m = (
                self._design.get("mooring_chain_density", 19900)
                * chain_diameter**2
            )  # kg/m
            rope_mass_per_m = (
                self._design.get("mooring_rope_density", 797.8)
                * rope_diameter**2
            )  # kg/m

            self.line_mass = (
                self.chain_length * chain_mass_per_m
                + self.rope_length * rope_mass_per_m
            ) / 1e3  # tonnes

        elif self.mooring_type == "Tlp":

            self.line_length = self.depth - draft

            self.line_mass = self.line_length * self.line_mass_per_m

        else:

            self.line_length = (
                0.0002 * (self.depth**2) + 1.264 * self.depth + 47.776 + fixed
            )

            self.line_mass = self.line_length * self.line_mass_per_m

    def calculate_anchor_mass_cost(self):
        """
        Returns the mass and cost of anchors.

        TODO: Anchor masses are rough estimates based on [1]. Should be
        revised when this module is overhauled in the future.
        TODO: Mooring types for Catenary, TLP, SemiTaut will likely have
        different anchors.
        """

        if self.mooring_type == "Semitaut":

            if self.anchor_type == "Drag Embedment":
                self.anchor_mass = 20

                # Interpolation of anchor cost at project depth
                self.anchor_cost = interp1d(
                    self._semitaut_params["depths"],
                    self._semitaut_params["anchor_costs"],
                    fill_value="extrapolate",
                )(self.depth).item()

            else:
                self.anchor_mass = 50
                self.anchor_cost = (
                    sqrt(self.breaking_load / 9.81 / 1250) * 150000
                )

        else:

            if self.anchor_type == "Drag Embedment":
                self.anchor_mass = 20
                self.anchor_cost = self.breaking_load / 9.81 / 20.0 * 2000.0

            else:
                self.anchor_mass = 50
                self.anchor_cost = (
                    sqrt(self.breaking_load / 9.81 / 1250) * 150000
                )

    @property
    def line_cost(self):
        """Returns cost of one line mooring line."""

        if self.mooring_type == "Semitaut":
            # Interpolation of line cost at project depth
            line_cost = interp1d(
                self._semitaut_params["depths"],
                self._semitaut_params["total_line_costs"],
                fill_value="extrapolate",
            )(self.depth).item()

        else:

            line_cost = self.line_length * self.line_cost_rate

        return line_cost

    @property
    def total_cost(self):
        """Returns the total cost of the mooring system."""

        return (
            self.num_lines
            * self.num_turbines
            * (self.anchor_cost + self.line_cost)
        )

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        return {
            "num_lines": self.num_lines,
            "line_diam": self.line_diam,
            "line_mass": self.line_mass,
            "line_length": self.line_length,
            "line_cost": self.line_cost,
            "mooring_type": self.mooring_type,
            "anchor_type": self.anchor_type,
            "anchor_mass": self.anchor_mass,
            "anchor_cost": self.anchor_cost,
            "system_cost": self.total_cost,
        }

    @property
    def design_result(self):
        """Returns the results of the design phase."""

        return {"mooring_system": self.detailed_output}



"""Provides the base `DesignPhase` class."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from abc import abstractmethod

from wisdem.orbit.phases import BasePhase
from wisdem.orbit.core.defaults import common_costs


class DesignPhase(BasePhase):
    """BasePhase subclass for design modules."""

    expected_config = None
    output_config = None

    @property
    @abstractmethod
    def design_result(self):
        """
        Returns result of DesignPhase to be passed into config and consumed by
        InstallPhase.

        Returns
        -------
        dict
            Dictionary of design results.
        """

        return {}

    def get_default_cost(self, design_name, key, subkey=None):
        """Return the cost value for a key in a design
        dictionary read from common_cost.yaml.
        """

        if (design_dict := common_costs.get(design_name, None)) is None:
            raise KeyError(f"No {design_name} in common_cost.yaml.")

        if (cost_value := design_dict.get(key, None)) is None:
            raise KeyError(f"{key} not found in [{design_name}] common_costs.")

        if isinstance(cost_value, dict):
            if subkey is None:
                raise ValueError(
                    f"{key} is a dictionary and requires a 'subkey' input."
                )

            if (sub_cost_value := cost_value.get(subkey, None)) is None:
                raise KeyError(
                    f"{subkey} not found in [{design_name}][{cost_value}]"
                    " common_costs."
                )

            return sub_cost_value

        else:
            return cost_value



import os

import numpy as np

from wisdem.commonse.mpi_tools import MPI
from wisdem.glue_code.runWISDEM import run_wisdem

## File management
run_dir = os.path.dirname(os.path.realpath(__file__)) + os.sep
fname_wt_input = os.path.join(run_dir, "IEA15MW_FB_scaled.yaml")
fname_modeling_options = os.path.join(run_dir, "modeling_options.yaml")
fname_analysis_options = os.path.join(run_dir, "analysis_options.yaml")

wt_opt, modeling_options, analysis_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)



"""Default inputs used throughout ORBIT."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

from pathlib import Path

import yaml

from wisdem.orbit.core.library import loader

DIR = Path(__file__).parent

with (DIR / "process_times.yaml").open() as f:
    process_times = yaml.load(f, Loader=loader)


with (DIR / "common_costs.yaml").open() as f:
    common_costs = yaml.load(f, Loader=loader)
